import cognitojwt
import logging
from cognitojwt import jwt_sync
import boto3
import traceback
import json
from flask import request
from functools import wraps
from flask_api import status as http_status
from cryptography.fernet import Fernet
from .connect_database import ConnectDatabase
from .response_json import Response

logging = logging.getLogger('app.sub')


class Library(object):
    def __init__(self, cognito_region='eu-west-1',
                 sql_settings=None,
                 standard_roles=None,
                 mongo_url=None,
                 app_client_id=None):
        # Declaring global variables

        self.app_client_id = app_client_id                   # app_client_id of AWS cognito
        self.mongo_url = mongo_url                           # Mongo database url
        self.standard_roles = standard_roles                 # standard roles initialised
        self.cognito_region = cognito_region                 # Aws cognito region
        self.data_token = None                               # data_token it is request payload
        self.username = None                                 # cognito username
        self.sql_settings = sql_settings                     # sql database configurations
        self.role = None                                     # user role in organisation
        self.role_permissions = {}                           # User role permissions set
        self.message_access_denied = "access_denied"         # return message for access denied
        self.process_failed_message = "failed_to_process"    # return message for process failure
        self.user_access_permissions = None                 # user permissions for respected tasks
        self.response_obj = Response()                       # template for return response
        self.connect_db_obj = ConnectDatabase(
            self.sql_settings, self.mongo_url)               # creating object for database connection

    def get_ssm_param(self, paramname, ssm_client):
        """
        Info: It hits aws ssm parameter store and returns the requested parameter value.
        param paramname: paramname should be the ssm parameter name.
        return: It returns ssm value which is assigned to ssm parameter.

        """
        try:
            session = boto3.Session(region_name=self.cognito_region)
            ssm = session.client(ssm_client)
            paramdata = ssm.get_parameter(Name=paramname, WithDecryption=True)
            return self.response_obj.output_response(data=format(paramdata['Parameter']['Value']))
        except Exception as error:
            return self.response_obj.output_response(error=False, status=500, data=str(error))

    def user_authorisation(self, userpool_id, testmode, type="Full"):
        """
        Info: It validate cognito token valididty and user existence
        :param userpool_id: cognito userpool id
        :param testmode: Boolean, If it is True then validation
        ignores token expiration else check for token expiration
        :return:
        """
        def authorisation_check(f):
            """
            Info: Decorator which authorise aws token
            :param f: Function
            :return:
            """
            @wraps(f)
            def decorated_function(*args, **kws):

                # Extracting header from the request
                if type == "full" or type == "aws" :
                    logging.info("getting token from Header")
                    token = request.headers['Authorization']

                    # Making sure that request token is not None

                    if token and not None:

                        # Calling aws_authorisation function for validating aws token

                        logging.info("processing aws authorisation")
                        aws_authorise = Library.aws_authorisation(
                            self, token, userpool_id, testmode)
                        logging.info("aws authorisation completed")

                        # Making sure that the function not returned any error
                        # aborting the process if any error occurs

                        if aws_authorise[self.response_obj.k_status] != 200:
                            return aws_authorise, aws_authorise[self.response_obj.k_status]
                        logging.info("user validation started by using"
                                     " account_id and cognito username")
                        if type == "aws":
                            return f(*args, **kws)

                    if type == "full" or type == "account":

                        # calling valid_user_check function to validate given aws
                        # token and account_id referring same person or not

                        valid_user = Library.valid_user_check(self)
                        logging.info("user validation completed")

                        # Checking whether given user valid or not

                        if valid_user[self.response_obj.k_data] is True:
                            logging.info("found user is valid processing further")
                            return f(*args, **kws)

                        # If user not valid then aborting the process by giving return message

                        else:
                            logging.info(valid_user)
                            logging.info("invalid user, aborting the process")
                            return valid_user, valid_user[self.response_obj.k_status]

                # The request token is None, So aborting process

                else:
                    logging.info("empty token found")
                    return self.response_obj.output_response(
                        error=True,
                        status=http_status.HTTP_403_FORBIDDEN,
                        message="authorization_header_token_is_invalid"), http_status.HTTP_403_FORBIDDEN

            return decorated_function

        return authorisation_check

    def aws_authorisation(self, token, userpool_id, testmode):
        """
        Info: Validate AWS token is valid or not
        :param token: Aws token
        :param userpool_id: AWS user pool id
        :param testmode: Boolean, If it is True then validation
        ignores token expiration else check for token expiration
        :return:
        """
        try:

            # Validating aws token by using jwt_sync library

            logging.info("aws authorisation started")
            data = jwt_sync.decode(token, self.cognito_region,
                                   userpool_id, self.app_client_id,
                                   testmode=testmode)

            # Getting username from the data after decoding aws token

            self.username = data['username']

            # returning status 200

            return self.response_obj.output_response()

        # Catching aws token expiry exception

        except cognitojwt.exceptions.CognitoJWTException as error:
            logging.info("exception occurred while processing "
                         "aws authorisation : {}".format(error))
            return self.response_obj.output_response(
                error=True,
                status=http_status.HTTP_401_UNAUTHORIZED,
                message="authorization_header_token_expired",
                data=str(error))

        # Catching any exception occurs in the process

        except Exception as error:
            logging.info("exception occurred while processing "
                         "aws authorisation : {}".format(error))
            return self.response_obj.output_response(
                error=True,
                status=http_status.HTTP_500_INTERNAL_SERVER_ERROR,
                message="authorization_header_token_is_invalid",
                data=str(error))

    def valid_user_check(self):
        """
        Info: Validate whether cognito username is belong to the account_id provided in payload
        :return: returns data True if authentication success
        """
        try:
            # Extracting payload from the request

            self.data_token = json.loads(request.form["payload"])
            logging.info(self.data_token)

            # Authenticating user by using cognito username and account_id

            result = self.connect_db_obj.execute_procedure(
                'user_authentication',
                (self.username, self.data_token['account_id']),
                "select",
                True)

            # Making sure that execution of stored procedure not returned any error

            if result[self.response_obj.k_status] != 200:
                logging.info(result[self.response_obj.k_data])
                return result

            # Fetching result from procedure execution

            valid_user = result[self.response_obj.k_data][0][0]

            # procedure returns 1 if user is valid

            if valid_user == 1:
                # returning True if user is valid
                return self.response_obj.output_response(data=True)

            # User is invalid means cognito username and account_id are not related to one person

            else:
                logging.info("user is invalid, So aborting process")
                return self.response_obj.output_response(
                            error=True,
                            status=http_status.HTTP_403_FORBIDDEN,
                            message=self.message_access_denied)

        # Catching any exception occurs in the process

        except Exception as error:
            logging.info("exception occurred while validating user : {}"
                         "".format(str(traceback.format_exc())))
            return self.response_obj.output_response(
                error=True,
                status=http_status.HTTP_500_INTERNAL_SERVER_ERROR,
                message="authorization_header_token_is_invalid",
                data=str(error))

    def role_policy_check(self, mongo_org_db,
                          mongo_role_collection,
                          action: dict = None):

        def permission_check(f):
            @wraps(f)
            def decorated_function(*args, **kws):
                try:
                    # Getting role permissions

                    logging.info("getting role permissions")
                    response = Library.get_role_permissions(self, mongo_org_db, mongo_role_collection)

                    # Making sure that method not returns any error

                    if response[self.response_obj.k_status] != 200:
                        return response, response[self.response_obj.k_status]

                    # If action is None processing further

                    if action is None:
                        return f(*args, **kws)

                    #  Fetching user access permissions and view org permissions

                    if self.role_permissions != {} and action is not None:
                        logging.info("getting user access permissions")
                        response = Library.policy_check(
                            self, action)
                        if response[self.response_obj.k_status] != 200:
                            return response
                        logging.info("got user access permissions")
                        self.user_access_permissions = response[self.response_obj.k_data]

                    # Processing further if user have any of the permission

                    if self.user_access_permissions is not None or self.role == 'Admin':
                        return f(*args, **kws)

                    # User not have enough permissions so aborting process
                    else:
                        return self.response_obj.output_response(
                            error=True,
                            status=http_status.HTTP_403_FORBIDDEN,
                            message=self.message_access_denied), http_status.HTTP_403_FORBIDDEN

                # Catching any exception occurs in the process

                except Exception as error:
                    logging.info("exception occurred while checking "
                                 "role policies : {}".format(error))
                    return self.response_obj.output_response(
                        error=True,
                        status=http_status.HTTP_500_INTERNAL_SERVER_ERROR,
                        message=self.process_failed_message,
                        data=str(error)), http_status.HTTP_500_INTERNAL_SERVER_ERROR

            return decorated_function

        return permission_check

    def get_role_permissions(self, mongo_org_db, mongo_role_collection):
        """
        Info: Get role permissions
        :param mongo_org_db: Mongo database name where organisation data stored
        :param mongo_role_collection: Mongo collection name where organisation roles stored
        :return:
        """
        try:
            # Getting user role name based on user_id and org_id

            logging.info("getting role name")
            response = self.connect_db_obj.execute_procedure(
                'get_role_name',
                (self.data_token["user_id"],
                 self.data_token["org_id"]),
                "select",
                True)
            logging.info("got role name")

            # Making sure that procedure execution not resulted any error
            if response[self.response_obj.k_status] != 200:
                return response

            # If procedure returns 0 then there is a mismatch between user_id and org_id

            if len(response[self.response_obj.k_data]) == 0:
                logging.info("found there is a mismatch between user_id and org_id")
                return self.response_obj.output_response(
                    error=False,
                    status=http_status.HTTP_403_FORBIDDEN,
                    message=self.message_access_denied)

            # Fetching role name from procedure response

            self.role = response[self.response_obj.k_data][0][0]

            # Checking where role name is in the list of standard roles

            if self.role in self.standard_roles:

                # Fetching role permissions if the user role is not Admin
                if self.role != 'Admin':

                    # Fetching role permissions from mongo database

                    logging.info("fetching role permissions from mongo")
                    collection_roles = self.connect_db_obj.connect_mongo_db(
                        mongo_org_db, mongo_role_collection)
                    if collection_roles[self.response_obj.k_status] != 200:
                        return collection_roles
                    data = collection_roles[self.response_obj.k_data].find_one(
                        {"role_name": self.role})
                    self.role_permissions = data["role_permissions"]
                    logging.info(self.role_permissions)
                    logging.info("got role permissions")

            # User role is not in the standard roles list

            else:

                # Fetching role permissions from mongo database

                logging.info("fetching role permissions from mongo")
                collection_roles = self.connect_db_obj.connect_mongo_db(
                    mongo_org_db, mongo_role_collection)
                if collection_roles[self.response_obj.k_status] != 200:
                    return collection_roles
                data = collection_roles[self.response_obj.k_data].find_one(
                    {"org_id": self.data_token['org_id'], "role_name": self.role})
                self.role_permissions = data["role_permissions"]
                logging.info("got role permissions")

            # Returning status 200
            return self.response_obj.output_response()

        # Catching any exception occurs in the process and aborting process

        except Exception as error:
            logging.info("exception occured while getting role "
                         "permissions : {}".format(error))
            return self.response_obj.output_response(
                error=True,
                status=http_status.HTTP_500_INTERNAL_SERVER_ERROR,
                message=self.process_failed_message,
                data=str(error))

    def policy_check(self, action):

        """
        Info: Check the company policies and role policies and return common in both
        :param action: company policies json object
        :return: Common policies in both, json object
        """
        try:
            logging.info(action)
            policy_dict = {}
            k = "Company"
            if k in action.keys() and self.role_permissions.keys():
                permission_set = list(set(action[k]).intersection(set(self.role_permissions[k])))
                if len(permission_set) > 0:
                    policy_dict[k] = permission_set

            # Returning data by including policy dict as data
            return self.response_obj.output_response(data=policy_dict)
        
        # Catching if any exception occurs in the process
        except Exception as error:
            logging.info("exception occurred while running policy_check : {}"
                         "".format(error))
            return self.response_obj.output_response(
                error=True,
                status=http_status.HTTP_500_INTERNAL_SERVER_ERROR,
                message=self.process_failed_message,
                data=str(error))

    def get_users(self):
        """
        Info: Fetch organisation level users from the database
        :return: Users data
        """
        try:
            # Executing stored procedure to get users from database
            response = self.connect_db_obj.execute_procedure(
                'get_users',
                (self.data_token["org_id"],),
                'select', True)

            # Check for any error response
            if response[self.response_obj.k_status] != 200:
                return self.response_obj.output_response(
                    error=False,
                    status=http_status.HTTP_500_INTERNAL_SERVER_ERROR,
                    message=self.process_failed_message)

            # Preparing list of users from json object
            user_list = []
            for user in response[self.response_obj.k_data]:
                user = json.loads(user[0])
                if user["user_status"] == "Active":
                    user_list.append(user)

            # Returning users list
            return self.response_obj.output_response(data=user_list)

        # Catching any exception if occurs in the process
        except Exception as error:
            logging.info("exception occurred while fetching users : {}"
                         "".format(error))
            return self.response_obj.output_response(
                error=True,
                status=http_status.HTTP_500_INTERNAL_SERVER_ERROR,
                message=self.process_failed_message,
                data=str(error))

    @staticmethod
    def view_company_policy_check(f):
        @wraps(f)
        def decorated_function(*args, **kws):
            logging.info("check company policies")
            return f(*args, **kws)

        return decorated_function

    @staticmethod
    def encrypt(secrete_key, string):
        """
        Info: Encrypt given string using secrete key
        :param secrete_key: Secrete key for encoding and decoding script
        :param string: String to encode
        :return: Encoded string
        """
        key = secrete_key.encode(encoding='utf-8')
        cipher_suite = Fernet(key)
        encoded_text = cipher_suite.encrypt(string.encode(encoding='utf-8')).decode("utf-8")
        return encoded_text

    @staticmethod
    def company_profile_data():
        return {
            "org_id": "",
            "company_number": "",
            "company_profile_id": "",
            "company_profile_status": "",
            "assignment": [
                {
                    "profile_id": "",
                    "user_status": "",
                    "account_id": "",
                    "imported_user": ""
                }]}

    @staticmethod
    def company_data():
        return {"Jurisdiction": "",
                "TradingOnMarket": "",
                "DTR5Applies": "",
                "RegisteredOfficeAddress": "",
                "SailAddress": "",
                "IncorporationDate": "",
                "OriginatingRegistry": {
                    "country": ""
                },
                "Accounts": {
                    "next_made_up_to": "",
                    "next_due": ""
                },
                "ConfirmationStatement": {
                    "next_made_up_to": "",
                    "next_due": ""
                }
                }
